/********************************************************************
 * Author: Seokpyo Hong
 * Update: 2005-09-26
 * Module: Case Insensitive String
 * Comment:
 *
 ********************************************************************/

#ifndef __CIS_CISTRING_H__
#define __CIS_CISTRING_H__	1

#include <algorithm>
#include <stdexcept>
#include <string>

namespace CIS
{

	class cistring : public std::string
	{
	public:

		//  / Ҹ
		explicit cistring() {}
		cistring(const cistring& s, size_type pos = 0, size_type n = std::string::npos) : std::string(s, pos, n) {}
		cistring(const char* s, size_type n) : std::string(s, n) {}
		cistring(const char* s) : std::string(s) {}
		cistring(size_type n, char c) : std::string(n, c) {}
		template <class InputIterator>
		cistring(InputIterator begin, InputIterator end) : std::string(begin, end) {}
		~cistring() {}

		cistring(const std::string& s) : std::string(s) {}


		// Ҵ 
		cistring& operator=(const cistring& s)
		{
			if (&s != this)
				assign(s.begin(), s.end());
			return *this;
		}
		cistring& operator=(const char* s)
		{ assign(s, s + _length(s)); return *this; }
		cistring& operator=(char c)
		{ assign(static_cast<size_type>(1), c); return *this; }


		//  Լ
		int compare(const cistring& s) const
		{ return _compare(data(), data() + size(), s.data(), s.data() + s.size()); }
		int compare(size_type pos1, size_type n1, const cistring& s) const
		{
			if (pos1 > size())
				_throw_out_of_range();
			return _compare(data() + pos1, data() + pos1 + std::min(n1, size() - pos1), s.data(), s.data() + s.size());
		}
		int compare(size_type pos1, size_type n1, const cistring& s, size_type pos2, size_type n2) const
		{
			if (pos1 > size() || pos2 > s.size())
				_throw_out_of_range();
			return  _compare(data() + pos1,
					data() + pos1 + std::min(n1, size() - pos1),
					s.data() + pos2,
					s.data() + pos2 + std::min(n2, size() - pos2));
		}
		int compare(const char* s) const
		{ return _compare(data(), data() + size(), s, s + _length(s)); }
		int compare(size_type pos1, size_type n1, const char* s, size_type n2 = std::string::npos) const
		{
			if (pos1 > size())
				_throw_out_of_range();
			return  _compare(data() + pos1,
					data() + pos1 + std::min(n1, size() - pos1),
					s, s + n2);
		}


		void _throw_out_of_range() const
		{ throw(std::out_of_range("cistring")); }


	public:

		static size_t _length(const char* s)
		{
			size_t i;
			for (i = 0; s[i]; ++i)
				;
			return i;
		}

		static bool _eq(char c1, char c2)
		{ return toupper(c1) == toupper(c2); }

		static bool _lt(char c1, char c2)
		{ return toupper(c1) < toupper(c2); }

		static int _cmp(const char* s1, const char* s2, size_t n)
		{
			for (size_t i = 0; i < n; ++i)
				if (!_eq(s1[i], s2[i]))
					return _lt(s1[i], s2[i]) ? -1 : 1;
			return 0;
		}

		static int _compare(const char* f1, const char* l1, const char* f2, const char* l2)
		{
			const ptrdiff_t n1 = l1 - f1;
			const ptrdiff_t n2 = l2 - f2;
			const int cmp = _cmp(f1, f2, std::min(n1, n2));
			return cmp != 0 ? cmp : (n1 < n2 ? -1 : (n1 > n2 ? 1 : 0));
		}

		static const char* _find(const char* s, size_t n, char c)
		{
			for ( ; n > 0 ; ++s, --n)
				if (_eq(*s, c))
					return s;
			return 0;
		}

	};

	// ==  Լ
	inline bool operator==(const cistring& s1, const cistring& s2)
	{ return s1.size() == s2.size() && cistring::_cmp(s1.data(), s2.data(), s1.size()) == 0; }
	inline bool operator==(const cistring& s1, const std::string& s2)
	{ return s1.size() == s2.size() && cistring::_cmp(s1.data(), s2.data(), s1.size()) == 0; }
	inline bool operator==(const std::string& s1, const cistring& s2)
	{ return s1.size() == s2.size() && cistring::_cmp(s1.data(), s2.data(), s1.size()) == 0; }
	inline bool operator==(const cistring& s1, const char* s2)
	{ size_t n = cistring::_length(s2); return n == s1.size() && cistring::_cmp(s2, s1.data(), n) == 0; }
	inline bool operator==(const char* s1, const cistring& s2)
	{ size_t n = cistring::_length(s1); return n == s2.size() && cistring::_cmp(s1, s2.data(), n) == 0; }

	// !=  Լ
	inline bool operator!=(const cistring& s1, const cistring& s2)
	{ return !(s1 == s2); }
	inline bool operator!=(const cistring& s1, const std::string& s2)
	{ return !(s1 == s2); }
	inline bool operator!=(const std::string& s1, const cistring& s2)
	{ return !(s1 == s2); }
	inline bool operator!=(const cistring& s1, const char* s2)
	{ return !(s1 == s2); }
	inline bool operator!=(const char* s1, const cistring& s2)
	{ return !(s1 == s2); }

	// <  Լ
	inline bool operator<(const cistring& s1, const cistring& s2)
	{ return cistring::_compare(s1.data(), s1.data() + s1.size(), s2.data(), s2.data() + s2.size()) < 0; }
	inline bool operator<(const cistring& s1, const std::string& s2)
	{ return cistring::_compare(s1.data(), s1.data() + s1.size(), s2.data(), s2.data() + s2.size()) < 0; }
	inline bool operator<(const std::string& s1, const cistring& s2)
	{ return cistring::_compare(s1.data(), s1.data() + s1.size(), s2.data(), s2.data() + s2.size()) < 0; }
	inline bool operator<(const cistring& s1, const char* s2)
	{ size_t n = cistring::_length(s2); return cistring::_compare(s2, s2 + n, s1.data(), s1.data() + s1.size()) < 0; }
	inline bool operator<(const char* s1, const cistring& s2)
	{ size_t n = cistring::_length(s1); return cistring::_compare(s1, s1 + n, s2.data(), s2.data() + s2.size()) < 0; }

	// >  Լ
	inline bool operator>(const cistring& s1, const cistring& s2)
	{ return s2 < s1; }
	inline bool operator>(const cistring& s1, const std::string& s2)
	{ return s2 < s1; }
	inline bool operator>(const std::string& s1, const cistring& s2)
	{ return s2 < s1; }
	inline bool operator>(const cistring& s1, const char* s2)
	{ return s2 < s1; }
	inline bool operator>(const char* s1, const cistring& s2)
	{ return s2 < s1; }

	// <=  Լ
	inline bool operator<=(const cistring& s1, const cistring& s2)
	{ return !(s2 < s1); }
	inline bool operator<=(const cistring& s1, const std::string& s2)
	{ return !(s2 < s1); }
	inline bool operator<=(const std::string& s1, const cistring& s2)
	{ return !(s2 < s1); }
	inline bool operator<=(const cistring& s1, const char* s2)
	{ return !(s1 < s1); }
	inline bool operator<=(const char* s1, const cistring& s2)
	{ return !(s2 < s1); }

	// >=  Լ
	inline bool operator>=(const cistring& s1, const cistring& s2)
	{ return !(s1 < s2); }
	inline bool operator>=(const cistring& s1, const std::string& s2)
	{ return !(s1 < s2); }
	inline bool operator>=(const std::string& s1, const cistring& s2)
	{ return !(s1 < s2); }
	inline bool operator>=(const cistring& s1, const char* s2)
	{ return !(s1 < s2); }
	inline bool operator>=(const char* s1, const cistring& s2)
	{ return !(s1 < s2); }

} // end of namespace CIS

#endif
